@@ -1,6 +1,5 @@
 /* gRun - GTK application launcher dialog
  * Copyright (C) 1998 Southern Gold Development <tangomanrulz@geocities.com>
- * All Win32 code Copyright (C) 1998 Troy Engel <tengel@sonic.net> 
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,34 +20,25 @@
 #include <gdk/gdkkeysyms.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h>
+#include <locale.h>
 
-#ifdef WIN32
-#include <windows.h>
-#define PATH_CHAR	";"
-#define DIR_CHAR	"\\"
-#define DIR_INT	'\\'
-#define DOT_CHAR	"_"
-#define ARG_BASE	0
-#define VERSION 	"0.9.2"
-#define PREFIX	"C:\\WINDOWS"
-
-#else
 #include "config.h"
-#include <strings.h>
 #include <unistd.h>
 #include <dirent.h>
 #include "grun2.xpm"
+
 #if defined (HAVE_GETTEXT) || defined (HAVE_CATGETS)
 #include <libintl.h>
 #else
 #include "intl/libintl.h"
 #endif
+
 #define PATH_CHAR	":"
 #define DIR_CHAR	"/"
 #define DIR_INT 	'/'
 #define DOT_CHAR	"."
 #define ARG_BASE	1
-#endif
 
 #include <sys/stat.h>
 
@@ -65,13 +55,29 @@ struct grun {
 	guchar status;
 };
 
-#ifdef WIN32
-# define gettext(Msgid) (Msgid)
-# define dgettext(Domainname, Msgid) (Msgid)
-# define dcgettext(Domainname, Msgid, Category) (Msgid)
-# define textdomain(Domainname) ((char *) Domainname)
-# define bindtextdomain(Domainname, Dirname) ((char *) Dirname)
-#endif /* WIN32 */
+/* Function prototypes */
+
+int isFileX(const gchar *file);
+gchar *getAssoc(const gchar *ext);
+
+char *getLine(FILE *file);
+int isFileExec(const gchar *file);
+GList *loadHistory();
+int saveHistory(const char *cmd, sgrun *gdat);
+void gquit();
+void startApp(const gchar *cmd, sgrun *gdat);
+gint gcomplete(sgrun *gdat, const gchar *twrk, gint oldpos);
+gint gdircomplete(sgrun *gdat, const gchar *twrk, gint oldpos);
+gint gdirmapcomplete(sgrun *gdat, const gchar *twrk, gint oldpos);
+gint gclick(GtkWidget *widget, GdkEventKey *event, gpointer data);
+gint gexit(GtkWidget *widget, GdkEvent *event, gpointer data);
+void launch(GtkWidget *widget, gpointer data);
+void cancel(GtkWidget *widget, gpointer data);
+void bclose(GtkWidget *widget, gpointer data);
+void bcancel(GtkWidget *widget, gpointer data);
+void browse(GtkWidget *widget, gpointer data);
+
+/* End prototypes */
 
 char *getLine(FILE *file) {
 	char *tmp, in;
@@ -105,8 +111,10 @@ char *getLine(FILE *file) {
 	}	
 }
 
-#ifdef TESTFILE
 int isFileX(const gchar *file) {
+#ifndef TESTFILE
+        return TRUE; // Ie, we'll consider everything an executable.
+#else
 	FILE *fHnd;
 	gchar *fname, *home_env, *line, *twrk, *hold, *split, *rdx;
 	struct stat buff;
@@ -126,7 +134,7 @@ int isFileX(const gchar *file) {
 	err = stat(fname, &buff);
 	if (err == -1) {
 			g_free(fname);
-			fname = g_strconcat(PREFIX, "/share/grun/consfile", NULL);
+			fname = g_strconcat(SYSCONFDIR, "/consfile", NULL);
 	}
 
 	fHnd = fopen(fname, "rb");
@@ -153,12 +161,13 @@ int isFileX(const gchar *file) {
 	g_free(hold);
 	fclose(fHnd);
 	return TRUE;
+#endif /* TESTFILE */
 }
-#endif
 
-#ifdef ASSOC
 gchar *getAssoc(const gchar *ext) {
-#ifndef WIN32
+#ifndef ASSOC
+  return NULL;
+#else
 	FILE *fHnd;
 	gchar *line, *split, *fname, *rsp, *home_env;
 	struct stat buff;
@@ -170,7 +179,7 @@ gchar *getAssoc(const gchar *ext) {
 	err = stat(fname, &buff);
 	if (err == -1) {
 			g_free(fname);
-			fname = g_strconcat(PREFIX, "/share/grun/gassoc", NULL);
+			fname = g_strconcat(SYSCONFDIR, "/gassoc", NULL);
 	}
 
 	fHnd = fopen(fname, "rb");
@@ -195,11 +204,8 @@ gchar *getAssoc(const gchar *ext) {
 	}
 	fclose(fHnd);
 	return NULL;
-#else
-/* Win lookup code goes here. Possibly HKey lookup? */
-#endif
+#endif /* ASSOC */
 }
-#endif
 
 int isFileExec(const gchar *file) {
 	int err;
@@ -208,55 +214,34 @@ int isFileExec(const gchar *file) {
 
 	inc = g_strdup(file);
 	hold = inc;
-#ifndef WIN32
 	split = strsep(&inc, " ");
-#else
-	split = strtok(inc, " ");
-#endif
 
-	err = stat(split, &buff);
-	if (err == -1) {
-		twrk = getenv("PATH");
-		path = g_strdup(twrk);
-		inc = path;
-#ifndef WIN32
-		twrk = strsep(&path, PATH_CHAR);
-#else
-		twrk = strtok(path, PATH_CHAR);
-#endif
-		while (twrk) {
-			awrk = g_strconcat(twrk, DIR_CHAR, split, NULL);
-			err = stat(awrk, &buff);
-			g_free(awrk);
-			if (err != -1) {
-				g_free(hold);
-				g_free(inc);
-				if (buff.st_mode & S_IEXEC) {
-					return TRUE;
-				}
-				else {
-					return FALSE;
-				}
-			}
-			else {
-#ifndef WIN32
-				twrk = strsep(&path, PATH_CHAR);
-#else
-				twrk = strtok(NULL, PATH_CHAR);
-#endif
-			}
-		}
+	if (!stat(split, &buff)) {
+	  g_free(hold);
+	  return buff.st_mode & S_IEXEC ? TRUE : FALSE;
 	}
-	else {
-		g_free(hold);
-		if (buff.st_mode & S_IEXEC) {
-			return TRUE;
-		}
-		else {
-			return FALSE;
-		}
+
+	twrk = getenv("PATH");
+	path = g_strdup(twrk);
+	inc = path;
+	twrk = strsep(&path, PATH_CHAR);
+	
+	while (twrk) {
+	  awrk = g_strconcat(twrk, DIR_CHAR, split, NULL);
+	  err = stat(awrk, &buff);
+	  g_free(awrk);
+
+	  if (!err) {
+	    g_free(hold);
+	    g_free(inc);
+
+	    return buff.st_mode & S_IEXEC ? TRUE : FALSE;
+	  }
+
+	  twrk = strsep(&path, PATH_CHAR);
 	}
-	return -1;
+	
+	return -1; // File not found anywhere.
 }
 
 GList *loadHistory() {
@@ -327,39 +312,74 @@ void gquit() {
 	return;
 }
 
+/** Tries to fork and exec the command line passed as WORK.
+ * The command is executed as a grandchild of the current
+ * process. The intermediate child process exits immediately.
+ * This function returns only in the original parent process.
+ */
+void do_fork(const gchar * cmdline) {
+        gchar ** args;
+	gchar * twrk, * work;
+	int cnt, len, scnt;
+	gint pid = fork();
+
+	if (pid != 0) { // The main proces returns immediately
+	  return;
+	}
+
+	pid = fork();
+
+	if (pid == 0) {
+	  work = g_strdup(cmdline);
+	  len = strlen(work);
+	  scnt = 1;
+	  for (cnt = 0; cnt < len; cnt++) {
+	    if (work[cnt] == ' ') {
+	      scnt++;
+	    }
+	  }
+	  args = g_malloc(sizeof(char *) * (scnt + 1));
+	  args[scnt] = NULL;
+	  if (scnt == 1) {
+	    args[0] = g_strdup(work);
+	  }
+	  else {
+	    twrk = (char *) strsep(&work, " ");
+	    args[0] = twrk;
+	    cnt = 1;
+	    while (twrk) {
+	      twrk = (char *) strsep(&work, " ");
+	      args[cnt] = twrk;
+	      cnt++;
+	    }
+	    args[cnt] = work;
+	  }
+	  execvp(args[0], args);
+	}
+	
+	// The child process is just a helper and exits immediately.
+	// If the grandchild reaches here, it's because execvp failed,
+	// and so it exits, too.
+	_exit(0); 
+	
+}
+
 void startApp(const gchar *cmd, sgrun *gdat) {
-#ifndef WIN32
 	char **args, *work, *twrk, *assoc, *hold, *split;
-	int cnt, len, scnt, pid, res;
+	int cnt, len, scnt, res;
 
 	res = isFileExec(cmd);
-	if (res == -1) {
-			if (gdat) {
-				if (!(gdat->status & 0x80)) {
-					g_list_free(gdat->history);
-					gtk_widget_destroy(gdat->win);
-				}
-			}
-			else {
-				exit(0);
-			}
-	}
-	else {
+	if (res != -1) {
 		if (res) {
 #ifdef TESTFILE
-			res = isFileX(cmd);
-			if (res) {
-				work = g_strdup(cmd);
-			}
-			else {
-				work = g_strconcat(XTERM, " -e ", cmd, NULL);
-			}
+		  work =  isFileX(cmd) ? 
+		    g_strdup(cmd) : 
+		    g_strconcat(XTERM, " -e ", cmd, NULL);
 #else
-			work = g_strdup(cmd);
+		  work = g_strdup(cmd);
 #endif
 		}
 		else {
-#ifdef ASSOC
 			twrk = g_strdup(cmd);
 			hold = twrk;
 			split = strsep(&twrk, " ");
@@ -374,181 +394,67 @@ void startApp(const gchar *cmd, sgrun *gdat) {
 					work = g_strconcat(assoc, " ", cmd, NULL);
 					g_free(assoc);
 				}
-				else {
-					if (gdat) {
-						if (!(gdat->status & 0x80)) {
-							g_list_free(gdat->history);
-							gtk_widget_destroy(gdat->win);
-						}
-					}
-					else {
-						exit(0);
-					}
-				}
 			}
-			else {
-				if (gdat) {
-					if (!(gdat->status & 0x80)) {
-						g_list_free(gdat->history);
-						gtk_widget_destroy(gdat->win);
-					}
-				}
-				else {
-					exit(0);
-				}
-			}
-#else
-			if (gdat) {
-				if (!(gdat->status & 0x80)) {
-					g_list_free(gdat->history);
-					gtk_widget_destroy(gdat->win);
-				}
-			}
-			else {
-				exit(0);
-			}
-#endif
 		}
 	}
 
-	pid = fork();
-	if (pid == 0) {
-		pid = fork();
-		if (pid == 0) {
-			len = strlen(work);
-			scnt = 1;
-			for (cnt = 0; cnt < len; cnt++) {
-				if (work[cnt] == ' ') {
-					scnt++;
-				}
-			}
-			args = g_malloc(sizeof(char *) * (scnt + 1));
-			args[scnt] = NULL;
-			if (scnt == 1) {
-					args[0] = g_strdup(work);
-			}
-			else {
-				twrk = (char *) strsep(&work, " ");
-				args[0] = twrk;
-				cnt = 1;
-				while (twrk) {
-					twrk = (char *) strsep(&work, " ");
-					args[cnt] = twrk;
-					cnt++;
-				}
-				args[cnt] = work;
-			}
-			execvp(args[0], args);
-			_exit(0);
-		}
-		else {
-			_exit(0);
-		}
+	if (work) {
+	  do_fork(work);
 	}
-	else {
-		if (gdat) {
-			cnt = saveHistory(cmd, gdat);
-			if (gdat->status & 0x80) {
-				if (cnt) {
-					gdat->history = g_list_prepend(gdat->history, (gpointer) cmd);
-					gtk_combo_set_popdown_strings(GTK_COMBO (gdat->cmb), gdat->history);
-				}
-				gtk_entry_select_region(GTK_ENTRY (GTK_COMBO (gdat->cmb)->entry), 0, -1);
-				gtk_entry_set_position(GTK_ENTRY (GTK_COMBO (gdat->cmb)->entry), 0);
-				gdat->cmdLen = 0;
-				gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
-			}
-			else {
-				g_list_free(gdat->history);
-				gtk_widget_destroy(gdat->win);
-			}
-		}
-		else {
-			_exit(0);
-		}
-	}
-#else /* Win32 */
-	STARTUPINFO startupinfo;
-	PROCESS_INFORMATION processinfo;
-	int cnt;
 
-	startupinfo.cb = sizeof (STARTUPINFO);
-	startupinfo.lpReserved = NULL;
-	startupinfo.lpDesktop = NULL;
-	startupinfo.lpTitle = NULL;
-	startupinfo.dwFlags = STARTF_USESHOWWINDOW;
-	startupinfo.wShowWindow = SW_SHOWDEFAULT;
-	startupinfo.cbReserved2 = 0;
-	startupinfo.lpReserved2 = NULL;
+	if (!gdat) {
+	  _exit(0);
+	}
 
-	/* this is implemented using NULL for the app name
-	   and the entire cmdline in the lpCmdLine param on
-	   purpose. WinNT firing off 16bit apps requires
-	   CreateProcess to be called this way. TODO: parse
-	   the cmdline and figure out if it's a 16bit app
-	   on our own, then adjust everything.
-	*/
-	/* TODO: doesn't handle files, folders and .lnk like
-	   shellexecute does. fix it. */
-	if (!CreateProcess (NULL, (gchar *)cmd, NULL, NULL,
-			TRUE, NORMAL_PRIORITY_CLASS, NULL,
-			NULL, &startupinfo, &processinfo)) {
-		g_message ("CreateProcess failed\n");
-		_exit(0);
+	cnt = saveHistory(cmd, gdat);
+
+	if ( ! (gdat->status & 0x80)) {
+	  g_list_free(gdat->history);
+	  gtk_widget_destroy(gdat->win);
+	  return;
 	}
-	CloseHandle (processinfo.hThread);
-	if (gdat) { /* gdat = NULL if cmdline launched */
-		cnt = saveHistory(cmd, gdat);
-		if (gdat) {
-			if (gdat->status & 0x80) {
-				if (cnt) {
-					gdat->history = g_list_prepend(gdat->history, (gpointer) cmd);
-					gtk_combo_set_popdown_strings(GTK_COMBO (gdat->cmb), gdat->history);
-				}
-				gtk_entry_select_region(GTK_ENTRY (GTK_COMBO (gdat->cmb)->entry), 0, -1);
-				gtk_entry_set_position(GTK_ENTRY (GTK_COMBO (gdat->cmb)->entry), 0);
-				gdat->cmdLen = 0;
-				gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
-			}
-			else {
-				g_list_free(gdat->history);
-				gtk_widget_destroy(gdat->win);
-			}
-		}
-		else {
-			exit(0);
-		}
+
+	if (cnt) {
+	  gdat->history = g_list_prepend(gdat->history, (gpointer) cmd);
+	  gtk_combo_set_popdown_strings(GTK_COMBO (gdat->cmb), gdat->history);
 	}
-#endif /* WIN32 */
+
+	/* the order does matter here;
+	   move the cursor first,
+	   then select the region */
+	gtk_entry_set_position(GTK_ENTRY (GTK_COMBO (gdat->cmb)->entry), 0);
+	gtk_entry_select_region(GTK_ENTRY (GTK_COMBO (gdat->cmb)->entry), 0, -1);
+	gdat->cmdLen = 0;
+	gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
 }
 
 /* This function takes a command fragment and searches through
    the history for the first match it finds. This is loaded into
    the entry box */
 
-gint gcomplete(sgrun *gdat, const gchar *twrk) {
+gint gcomplete(sgrun *gdat, const gchar *twrk, gint oldpos) {      
 		GList *work;
 		gint pos, len;
 
 		pos = strlen(twrk);
 		work = gdat->history;
+
 		while (work) {
-#ifdef WIN32 /* we're a case-insensative platform [sigh] */
-			if (strnicmp(twrk, (gchar *) work->data, pos) == 0) {
-#else
 			if (strncmp(twrk, (gchar *) work->data, pos) == 0) {
-#endif
 				gdat->cmdLen = pos;
 				gtk_entry_set_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), (gchar *) work->data);
 				len = strlen(work->data);
-				gtk_entry_select_region(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), pos, len);
 				gtk_entry_set_position(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), pos);
 				gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
 				return TRUE;
 			}
 			work = work->next;
 		}
-		return FALSE;
+
+                        /* if searching the history fails,
+                           then search the PATH
+                        */
+                        return gdircomplete(gdat, twrk, oldpos);
 }
 
 /* This function takes a command fragment and searches through
@@ -556,7 +462,7 @@ gint gcomplete(sgrun *gdat, const gchar *twrk) {
    that it finds. This is loaded into the entry box. This is bound
 	to Tab/Esc/F2 if the first character is not / */
 
-gint gdircomplete(sgrun *gdat, const gchar *twrk) {
+gint gdircomplete(sgrun *gdat, const gchar *twrk, gint oldpos) {
 	DIR *dir;
 	gchar *path, *inc, *wdir, *hold;
 	struct dirent *file;
@@ -568,32 +474,67 @@ gint gdircomplete(sgrun *gdat, const gchar *twrk) {
 	path = g_strdup(inc);
 	hold = path;
 	if (strlen(path) < 1) {
+          g_free(hold);
 		return FALSE;
 	}
-#ifdef WIN32
-	wdir = strtok(path, PATH_CHAR);
-#else
 	wdir = strsep(&path, PATH_CHAR);
-#endif
 	while (wdir) {
 		dir = opendir(wdir);
 		if (dir) {
 			file = readdir(dir);
 			while (file) {
-#ifdef WIN32 /* case insensative again */
-				if (strnicmp(twrk, file->d_name, pos) == 0) {
-#else
 				if (strncmp(twrk, file->d_name, pos) == 0) {
-#endif
 					inc = g_strconcat(wdir, DIR_CHAR, file->d_name, NULL);
 					err = stat(inc, &buf);
 					g_free(inc);
-					if (buf.st_mode & S_IEXEC) {
+					if ((err == 0) && (buf.st_mode & S_IEXEC)) {
 						gdat->cmdLen = pos;
 						gtk_entry_set_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), (gchar *) file->d_name);
 						len = strlen(file->d_name);
-						gtk_entry_select_region(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), pos, len);
 						gtk_entry_set_position(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), pos);
+                                                gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
+						closedir(dir);
+						g_free(hold);
+						return TRUE;
+					}
+				}
+				file = readdir(dir);
+			}
+			closedir(dir);
+		}
+		wdir = strsep(&path, PATH_CHAR);
+	}
+
+                g_free(hold);
+
+                /*
+                  discard the result of previous auto-completion
+                  if the user added a new character
+                */
+                if (pos <= oldpos)
+                  return FALSE;
+
+	inc = getenv("PATH");
+	path = g_strdup(inc);
+	hold = path;
+	if (strlen(path) < 1) {
+		return FALSE;
+	}
+	wdir = strsep(&path, PATH_CHAR);
+	while (wdir) {
+		dir = opendir(wdir);
+		if (dir) {
+			file = readdir(dir);
+			while (file) {
+				if (strncmp(twrk, file->d_name, oldpos) == 0) {
+					inc = g_strconcat(wdir, DIR_CHAR, file->d_name, NULL);
+					err = stat(inc, &buf);
+					g_free(inc);
+					if ((err == 0) && (buf.st_mode & S_IEXEC)) {
+                                          gdat->cmdLen = pos;
+                                          gtk_entry_set_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), twrk);
+                                          gtk_editable_set_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), -1);
+                                          gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
 						closedir(dir);
 						g_free(hold);
 						return TRUE;
@@ -603,11 +544,7 @@ gint gdircomplete(sgrun *gdat, const gchar *twrk) {
 			}
 			closedir(dir);
 		}
-#ifdef WIN32
-		wdir = strtok(NULL, PATH_CHAR);
-#else
 		wdir = strsep(&path, PATH_CHAR);
-#endif
 	}
 	g_free(hold);
 	return FALSE;
@@ -619,7 +556,7 @@ gint gdircomplete(sgrun *gdat, const gchar *twrk) {
 	the match is a directory, a / is appended. This is bound
 	to Tab/Esc/F2 if the first character is a / */
 
-gint gdirmapcomplete(sgrun *gdat, const gchar *twrk) {
+gint gdirmapcomplete(sgrun *gdat, const gchar *twrk, gint oldpos) {
 	DIR *dir;
 	gchar *path, *frag, *wdir, *hold, *inc, *whold;
 	struct dirent *file;
@@ -639,11 +576,13 @@ gint gdirmapcomplete(sgrun *gdat, const gchar *twrk) {
 	if (dir) {
 		file = readdir(dir);
 		while (file) {
-#ifdef WIN32 /* case insensative again */
-			if (strnicmp(frag, file->d_name, lenf) == 0) {
-#else
+                  /* ignore the directory "." and ".." */
+                  if ((strcmp(file->d_name, ".") == 0)
+                      || (strcmp(file->d_name, "..") == 0)) {
+                    file = readdir(dir);
+                    continue;
+                  }
 			if (strncmp(frag, file->d_name, lenf) == 0) {
-#endif
 				inc = g_strconcat(wdir, file->d_name, NULL);
 				err = stat(inc, &buf);
 				if (buf.st_mode & S_IFDIR) {
@@ -653,12 +592,14 @@ gint gdirmapcomplete(sgrun *gdat, const gchar *twrk) {
 					whold = g_strdup(inc);
 				}
 				g_free(inc);
-				if (err != -1) {
+                                if ((err == 0)
+                                    && ((buf.st_mode & S_IFDIR)
+                                        ||(buf.st_mode & S_IEXEC))) {
 					gdat->cmdLen = pos;
 					gtk_entry_set_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), (gchar *) whold);
 					len = strlen(whold);
-					gtk_entry_select_region(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), pos, len);
 					gtk_entry_set_position(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), pos);
+					gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
 					closedir(dir);
 					g_free(whold);
 					g_free(hold);
@@ -672,6 +613,53 @@ gint gdirmapcomplete(sgrun *gdat, const gchar *twrk) {
 		}
 		closedir(dir);
 	}	
+	g_free(hold);
+
+        /*
+          discard the result of previous auto-completion
+          if the user added a new character
+        */
+        if (pos <= oldpos)
+          return FALSE;
+
+	path = g_strdup(twrk);
+	hold = path;
+        path[oldpos] = '\0';
+	frag = rindex(path, DIR_INT);
+	frag[0] = '\0';
+	frag++;
+	lenf = strlen(frag);
+        if (lenf < 1) {
+          g_free(hold);
+          return FALSE;
+        }
+	wdir = g_strconcat(path, "/", NULL);
+	
+	dir = opendir(wdir);
+	if (dir) {
+		file = readdir(dir);
+		while (file) {
+			if (strncmp(frag, file->d_name, lenf) == 0) {
+				inc = g_strconcat(wdir, file->d_name, NULL);
+				err = stat(inc, &buf);
+				g_free(inc);
+                                if ((err == 0)
+                                    && ((buf.st_mode & S_IFDIR)
+                                        ||(buf.st_mode & S_IEXEC))) {
+                                          gdat->cmdLen = pos;
+                                          gtk_entry_set_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), twrk);
+                                          gtk_editable_set_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), -1);
+                                          gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
+					closedir(dir);
+					g_free(hold);
+					return TRUE;
+				}
+			}
+			file = readdir(dir);
+		}
+		closedir(dir);
+	}	
+
 	g_free(hold);
 	return FALSE;		
 }
@@ -680,6 +668,10 @@ gint gclick(GtkWidget *widget, GdkEventKey *event, gpointer data) {
 	gchar *cmd, *twrk, *tmp;
 	gint res;
 	sgrun *gdat;
+        gint selection_start;
+        gint selection_end;
+        gchar *selected_cmd;
+        gchar *entire_cmd;
 
 	gdat = (sgrun *) data;
 /* Fire off executable in command box */
@@ -693,15 +685,37 @@ gint gclick(GtkWidget *widget, GdkEventKey *event, gpointer data) {
 			return FALSE;
 			break;
 /* Handle backspace with autocomplete */
-#ifndef WIN32 /* this is a WinUI handled issue for us */
 		case GDK_BackSpace:
-			cmd = gtk_entry_get_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry));
-			if ((strlen(cmd) > 0) && (gdat->cmdLen > 0)) {
-				twrk = g_strdup(cmd);
-				gdat->cmdLen--;
-				twrk[gdat->cmdLen] = '\0';
+                  /* delete everything if the entire command is selected and
+                     if the cursor is at the last of the command
+                  */
+                  if (gtk_editable_get_selection_bounds(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), &selection_start, &selection_end))
+                  {
+                    selected_cmd = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), selection_start, selection_end);
+                    entire_cmd = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), 0, -1);
+                    if ((strcmp(selected_cmd, entire_cmd) == 0)
+                        && (gtk_editable_get_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry)) == selection_end))
+                    {
+                      gtk_entry_set_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), "");
+                      gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
+                      g_free(selected_cmd);
+                      g_free(entire_cmd);
+                      return TRUE;
+                    }
+                    g_free(selected_cmd);
+                    g_free(entire_cmd);
+                  }
+
+                  cmd = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), 0, gtk_editable_get_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry)));
+                  if ((strlen(cmd) > 0) && (gtk_editable_get_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry)) > 0)) {
+                    twrk = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), 0, gtk_editable_get_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry)) - 1);
 				if (strlen(twrk) > 0) {
-					res = gcomplete(gdat, twrk);
+			if (twrk[0] == DIR_INT) {
+                                    res = gdirmapcomplete(gdat, twrk, strlen(cmd));
+                                  }
+                                  else {
+					res = gcomplete(gdat, twrk, strlen(cmd));
+                                        }
 				}
 				else {
 					res = TRUE;
@@ -709,58 +723,73 @@ gint gclick(GtkWidget *widget, GdkEventKey *event, gpointer data) {
 					gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
 				}
 				g_free(twrk);
+                                g_free(cmd);
 				return res;
 			}
 			else {
+                          g_free(cmd);
 				return FALSE;
 			}
 			break;
-#endif
 /* Autocomplete from PATH if Tab */
-#ifndef WIN32 /* Win UI: TAB=next control, ESC quits */
-		case GDK_Tab:
 		case GDK_Escape:
-#else
-		case GDK_F2: /* arbitrary, F1 == Help in WinUI */
-#endif
-			cmd = gtk_entry_get_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry));
-			if ((strlen(cmd) == 0) || (gdat->cmdLen == 0)) {
+          cancel(widget,data);
+          return FALSE;
+          break;
+		case GDK_Tab:
+                  cmd = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), 0, gtk_editable_get_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry)));
+			if (strlen(cmd) == 0) {
 				return FALSE;
 			}
 			tmp = g_strdup(cmd);
 			twrk = g_strdup(cmd);
-			if ((gint)strlen(cmd) != gdat->cmdLen) {
-				twrk[gdat->cmdLen] = '\0';
-			}
-#ifdef WIN32
-			if ((twrk[0] == DIR_INT) || (twrk[1] == ':')) {
-#else
 			if (twrk[0] == DIR_INT) {
-#endif
-				res = gdirmapcomplete(gdat, twrk);
+				res = gdirmapcomplete(gdat, twrk, strlen(cmd));
 			}
 			else {
-				res = gdircomplete(gdat, twrk);
-			}
-			cmd = gtk_entry_get_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry));
-			if (strcmp(cmd, tmp) != 0) {
-					gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
+				res = gdircomplete(gdat, twrk, strlen(cmd));
 			}
 			g_free(twrk);
 			g_free(tmp);
+                        g_free(cmd);
 			return res;
 			break;
 /* Autocomplete from history */
 		default:
 			/* make sure it's a non-command key */
-			cmd = gtk_entry_get_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry));
+                  cmd = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), 0, gtk_editable_get_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry)));
 			if (event->length > 0) {
+                  /* replace everything if the entire command is selected and
+                     if the cursor is at the last of the command
+                  */
+                  if (gtk_editable_get_selection_bounds(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), &selection_start, &selection_end))
+                  {
+                    selected_cmd = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), selection_start, selection_end);
+                    entire_cmd = gtk_editable_get_chars(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), 0, -1);
+                    if ((strcmp(selected_cmd, entire_cmd) == 0)
+                        && (gtk_editable_get_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry)) == selection_end))
+                    {
+                      gtk_entry_set_text(GTK_ENTRY ((GTK_COMBO (gdat->cmb))->entry), event->string);
+                      gtk_editable_set_position(GTK_EDITABLE ((GTK_COMBO (gdat->cmb))->entry), -1);
+                      gtk_signal_emit_stop_by_name(GTK_OBJECT ((GTK_COMBO (gdat->cmb))->entry), "key_press_event");
+                      g_free(cmd);
+                      /* the input is appended to twrk below */
+                      cmd = g_strdup("");
+                      /* no return here; we need auto-completion */
+                    }
+                    g_free(selected_cmd);
+                    g_free(entire_cmd);
+                  }
 				twrk = g_strdup(cmd);
 				twrk = g_realloc(twrk, (strlen(twrk) + event->length + 1));
-				twrk[gdat->cmdLen] = '\0';
 				strncat(twrk, event->string, event->length);
 				if (strlen (twrk) > 0) {
-					res = gcomplete(gdat, twrk);
+			if (twrk[0] == DIR_INT) {
+                                    res = gdirmapcomplete(gdat, twrk, strlen(cmd));
+                                  }
+                                  else {
+					res = gcomplete(gdat, twrk, strlen(cmd));
+                                        }
 				}
 				else {
 					res = TRUE;
@@ -773,6 +802,7 @@ gint gclick(GtkWidget *widget, GdkEventKey *event, gpointer data) {
 					}
 				}
 				g_free(twrk);
+                                g_free(cmd);
 				return res;
 			}
 			else {
@@ -780,6 +810,7 @@ gint gclick(GtkWidget *widget, GdkEventKey *event, gpointer data) {
 				if ((event->keyval == GDK_End) && (strlen(cmd) != gdat->cmdLen)) {
 					gdat->cmdLen = strlen(cmd);
 				}
+                                g_free(cmd);
 				return FALSE;
 			}
 			break;
@@ -842,66 +873,22 @@ void browse(GtkWidget *widget, gpointer data) {
 	gtk_widget_show(rsp);
 }
 
-#ifndef WIN32
-void main(int argc, char **argv) {
-#else
-int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nShowCmd) {
-#endif
-	GtkWidget *win, *btn, *lbl, *cmb, *fix;
+int main(int argc, char **argv) {
+        GtkWidget *win, *btn, *lbl, *cmb, *main_box, *button_box;
 	GtkTooltips *tips;
 	int c, cnt, len, persist, tooltips;
 	char *home_env, *fname, *cmd, *icmd;
 	GList *list;
 	sgrun *gdat;
-#ifndef WIN32
 	GdkPixmap *pix;
 	GdkBitmap *mask;
 	GtkStyle *style;
-#endif
-
-#ifdef WIN32
-	/* parse into a command line ala argv, argc */
-	char **argv, *work, *twrk;
-	int argc;
-
-	len = strlen(lpCmdLine);
-	argc = 0;
-	for (cnt = 0; cnt < len; cnt++) {
-		if (lpCmdLine[cnt] == ' ') {
-			argc++;
-		}
-	}
-	argv = g_malloc(sizeof(char *) * (argc + 1));
-	argv[argc] = NULL;
-	if (argc == 1) {
-			argv[0] = g_malloc(sizeof(char) * (len + 1));
-			strcpy(argv[0], lpCmdLine);
-	}
-	else {
-		work = g_malloc(sizeof(char) * (len + 1));
-		strcpy(work, lpCmdLine);
-		twrk = (char *) strtok(work, " ");
-		argv[0] = twrk;
-		cnt = 1;
-		while (twrk) {
-			twrk = (char *) strtok(NULL, " ");
-			argv[cnt] = twrk;
-			cnt++;
-		}
-		argv[cnt] = work;
-	}
-#endif /* WIN32 */
 
 	gtk_init(&argc, &argv);
-#ifndef WIN32
 	setlocale (LC_ALL, "");
 	bindtextdomain (PACKAGE, LOCALEDIR);
 	textdomain (PACKAGE);
-#endif /* WIN32 */
  
-/* TODO: these don't do anything under Win32
-   because we've made it a non-console app.
-   Maybe use some sort of dialogbox. */
 /* New parameter scan routine, checks all parameters */
 	persist = FALSE;
 	tooltips = TRUE;
@@ -938,7 +925,6 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 		}
 			
 		if (argv[1][0] != '-') {
-#ifndef WIN32
 			len = 0;
 			for (c = 1; c < argc; c++) {
 				len += strlen(argv[c]) + 1;
@@ -950,9 +936,6 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 			}
 			cmd[len - 1] = '\0';
 			startApp(cmd, NULL);
-#else /* WIN32 */
-			startApp(lpCmdLine, NULL);
-#endif
 		}
 	}
 
@@ -969,7 +952,7 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 	sprintf(fname, gettext("gRun %s"), VERSION);
 	gtk_window_set_title(GTK_WINDOW(win), fname);
 	g_free(fname);
-	gtk_widget_set_usize(GTK_WIDGET(win), 250, 100);
+
 	gtk_signal_connect(GTK_OBJECT (win), "destroy", (GtkSignalFunc) gexit, NULL);
 	gtk_container_border_width(GTK_CONTAINER (win), 5);
 
@@ -981,12 +964,15 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 		gtk_tooltips_disable(tips);
 	}
 
-	fix = gtk_fixed_new();
-	gtk_container_add(GTK_CONTAINER(win), fix);
+	main_box = gtk_vbox_new(FALSE, 5);
+	gtk_container_add(GTK_CONTAINER(win), main_box);
 
 	lbl = gtk_label_new(gettext("Launch Application"));
-	gtk_fixed_put(GTK_FIXED (fix), lbl, 65, 4);
-	gtk_widget_show(lbl);
+        button_box = gtk_hbutton_box_new();
+
+	gtk_box_pack_start_defaults(GTK_BOX (main_box), lbl);
+	gtk_box_pack_end_defaults(GTK_BOX (main_box), button_box);
+
 
 	list = loadHistory();
 	gdat->history = list;
@@ -1001,8 +987,9 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 	cmb = gtk_combo_new();
 	gdat->cmb = cmb;
 	gdat->rsp = NULL;
-	gtk_widget_set_usize(GTK_WIDGET (GTK_COMBO(cmb)->entry), 215, 24);
-	gtk_fixed_put(GTK_FIXED (fix), cmb, 5, 27);
+
+	gtk_box_pack_start_defaults(GTK_BOX (main_box), cmb);
+
 	if (list) {
 		gtk_combo_set_popdown_strings(GTK_COMBO (cmb), list);
 		gtk_entry_set_text(GTK_ENTRY (GTK_COMBO (cmb)->entry), list->data);
@@ -1018,14 +1005,12 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 	}
 	gtk_signal_connect(GTK_OBJECT (GTK_COMBO (cmb)->entry), "key_press_event", (GtkSignalFunc) gclick, (gpointer) gdat);
 	gtk_window_set_focus(GTK_WINDOW (win), GTK_COMBO (cmb)->entry);
-	gtk_widget_show(cmb);
 
 	btn = gtk_button_new_with_label(gettext("OK"));
-	gtk_widget_set_usize(GTK_WIDGET(btn), 70, 24);
-	gtk_fixed_put(GTK_FIXED (fix), btn, 5, 60);
+
+	gtk_box_pack_start_defaults(GTK_BOX (button_box), btn);
 	gtk_signal_connect(GTK_OBJECT (btn), "clicked", (GtkSignalFunc) launch, (gpointer) gdat);
 	gtk_tooltips_set_tip(tips, btn, gettext("Launch application"), gettext("The OK button launches the application given in the entry box"));
-	gtk_widget_show(btn);
 
 	if (persist) {
 		btn = gtk_button_new_with_label(gettext("Close"));
@@ -1034,8 +1019,7 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 		btn = gtk_button_new_with_label(gettext("Cancel"));
 	}
 
-	gtk_widget_set_usize(GTK_WIDGET(btn), 70, 24);
-	gtk_fixed_put(GTK_FIXED (fix), btn, 85, 60);
+	gtk_box_pack_start_defaults(GTK_BOX (button_box), btn);
 	gtk_signal_connect(GTK_OBJECT (btn), "clicked", (GtkSignalFunc) cancel, (gpointer) gdat);
 	if (persist) {
 		gtk_tooltips_set_tip(tips, btn, gettext("Close gRun"), gettext("The Close button will close the persistant gRun window"));
@@ -1043,26 +1027,18 @@ int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nS
 	else {
 		gtk_tooltips_set_tip(tips, btn, gettext("Cancel launch"), gettext("The Cancel button cancels the launch process and closes gRun"));
 	}
-	gtk_widget_show(btn);
 
 	btn = gtk_button_new_with_label(gettext("Browse"));
-	gtk_widget_set_usize(GTK_WIDGET(btn), 70, 24);
-	gtk_fixed_put(GTK_FIXED (fix), btn, 165, 60);
+	gtk_box_pack_start_defaults(GTK_BOX (button_box), btn);
 	gtk_signal_connect(GTK_OBJECT (btn), "clicked", (GtkSignalFunc) browse, (gpointer) gdat);
 	gtk_tooltips_set_tip(tips, btn, gettext("Browse for command"), gettext("The Browse button opens a file selection box so the user can browse for the command to launch"));
-	gtk_widget_show(btn);
 
-	gtk_widget_show(fix);
 	gtk_window_set_policy(GTK_WINDOW (win), FALSE, FALSE, FALSE);
 	gtk_widget_realize(win);
-/* Icon, can this work for Win? */
-#ifndef WIN32
-	style = gtk_widget_get_style(win);
-	pix = gdk_pixmap_create_from_xpm_d(win->window, &mask, &style->bg[GTK_STATE_NORMAL], grun2);
-	gdk_window_set_icon(win->window, NULL, pix, mask);
-	g_free(pix);
-	g_free(mask);
-#endif
-	gtk_widget_show(win);
+	
+	gtk_window_set_icon(GTK_WINDOW(win), gdk_pixbuf_new_from_xpm_data(grun2));
+
+	gtk_widget_show_all(win);
 	gtk_main();
+	return 0;
 }
